from collections import deque
from typing import List, Union, Optional
from itertools import groupby

# Same type aliases
Lit = int
Var = int
Clause = List[Lit]
CNF = List[Clause]
Assignment = List[Optional[bool]]


# FOR DEBUGGING:
# Replace this with a CNF that breaks your solver
n = 2
cnf = [[-1, 2], [-1, -2]]


def preprocess(cnf: CNF) -> CNF:
    """ Remove duplicate literals and clauses from a CNF formula. """
    cnf = [list(set(clause)) for clause in cnf]
    cnf.sort()
    return list(clause for clause, _ in groupby(cnf))


class PennSAT:
    """A DPLL-based SAT solver with a static most-frequent decision heuristic."""

    def __init__(self, n: int, cnf: CNF, heuristic: bool = True, unit_propagation: bool = True):
        # The number of variables
        self.n = n
        # The CNF as a list of clauses
        self.cnf: CNF = preprocess(cnf)
        # A stack of partial truth assignments: lists mapping each variable to True/False/None
        self.assignment_stack: List[Assignment] = [[None] * (n + 1)]
        if heuristic:
            # We'll implement this function later
            self.var_ordering = self.compute_frequency_ordering()
        else:
            self.var_ordering = list(range(1, n + 1))
        # A flag indicating whether the solver should do unit propagation
        self.unit_propagation = unit_propagation
        # A stack of decision variables
        self.decision_stack: List[Var] = []
        # A map from literal to a list of clauses containing that literal
        self.clauses_containing: List[List[Clause]] = [list() for _ in range(2 * n + 1)]
        # A queue of literals which have been assumed `False`
        self.propagation_queue: deque = deque()
        # TODO (2.1): add each clause in self.cnf to the clauses_containing list of each of its literals

    def value(self, literal: Lit) -> Optional[bool]:
        """Returns the value of the literal (`True`/`False`/`None`) under the current assignment."""
        # TODO (2.2)
        # HINT: a layer of the assignment stack maps each variable (not each literal) to True/False/None
        pass

    def assume(self, literal: Lit) -> bool:
        """Assigns the literal to `True` in the current assignment. Returns `True` if the literal was
        successfully assumed (or if it was already `True`). Returns `False` if it was already False.
        """
        # TODO (2.3)
        pass

    def compute_frequency_ordering(self) -> List[Var]:
        """Returns a list of variables `[1..n]` sorted by descending frequency score
        based on the stored CNF."""
        cnf = self.cnf
        n = self.n
        # TODO (2.4)
        pass

    def pick_variable(self) -> Optional[Var]:
        """Returns the first unassigned variable in the stored variable ordering,
        or `None` if all variables have been assigned."""
        # TODO (2.5)
        pass

    def backtrack_and_assume_negation(self) -> None:
        """Undo the last decision and restore the previous partial assignment.
        Then assume the negation of the last decision variable."""
        # TODO (2.6)
        pass

    def unit_propagate_clause(self, clause: Clause) -> bool:
        """
        Accepts as input a clause, and performs unit propagation if the
        clause is unit (unsatisfied and containing 1 unassigned literal).
        Returns `False` if a conflict is detected (unsatisfied and 0
        unassigned literals); otherwise `True`.
        """
        # TODO (2.10)
        pass

    def unit_propagate(self) -> bool:
        """
        Calls `unit_propagate_clause` on the clauses containing literals popped from `propagation_queue`
        until the queue is empty. If we ever encounter a conflict, this returns `False`; otherwise `True`.
        If `self.unit_propagation` is `False`, simply checks for a conflict without doing any unit propagation.
        """
        if not self.unit_propagation:
            return not any(
                all(self.value(literal) is False for literal in clause)  # clause is unsatisfied
                for clause in self.cnf
            )
        else:
            # TODO (2.9)
            # NOTE: `unit_propagate_clause` may call `assume`, which adds to the propagation queue
            pass

    def solve(self) -> Union[str, List[Lit]]:
        """Return a satisfying assignment to the stored CNF, or return `'UNSAT'` if none exists."""
        for clause in self.cnf:
            if len(clause) == 1 and not self.assume(clause[0]):
                return "UNSAT"
        if not self.unit_propagate():
            return "UNSAT"
        # TODO (2.7)


# solver = PennSAT(n, cnf)
# print(solver.solve())
# solver.solve()
